#include "GtQuaternion.h"
#include "GtMatrix4.h"

const GtQuaternion GtQuaternion::IDENTITY(1.0f, 0.0f, 0.0f, 0.0f);
const float GtQuaternion::ms_fEpsilon = 1e-03f;

GtQuaternion::GtQuaternion(float _w, float _x, float _y, float _z) :
w(_w),
x(_x),
y(_y),
z(_z)
{
}

GtQuaternion::GtQuaternion(float fAngle, const GtPoint3f& axis)
{
	FromAngleAxis(fAngle, axis);
}

void GtQuaternion::SetW(float _w)
{
	w = _w;
}

void GtQuaternion::SetX(float _x)
{
	x = _x;
}

void GtQuaternion::SetY(float _y)
{
	y = _y;
}

void GtQuaternion::SetZ(float _z)
{
	z = _z;
}

float GtQuaternion::GetW() const
{
	return w;
}

float GtQuaternion::GetX() const
{
	return x;
}

float GtQuaternion::GetY() const
{
	return y;
}

float GtQuaternion::GetZ() const
{
	return z;
}

void GtQuaternion::SetValue(float _w, float _x, float _y, float _z)
{
	w = _w; x = _x; y = _y; z = _z;
}

void GtQuaternion::GetValue(float &_w, float &_x, float &_y, float &_z)
{
	_w = w; _x = x; _y = y; _z = z;
}

GtQuaternion GtQuaternion::operator + (const GtQuaternion& q) const
{
	return GtQuaternion(w + q.w, x + q.x, y + q.y, z + q.z);
}

GtQuaternion GtQuaternion::operator - (const GtQuaternion& q) const
{
	return GtQuaternion(w - q.w, x - q.x, y - q.y, z - q.z);
}

GtQuaternion GtQuaternion::operator - (void) const
{
	return GtQuaternion(-w, -x, -y, -z);
}

GtQuaternion GtQuaternion::operator * (const GtQuaternion& q) const
{
	return GtQuaternion
	(
		w * q.w - x * q.x - y * q.y - z * q.z,
		w * q.x + x * q.w + y * q.z - z * q.y,
		w * q.y + y * q.w + z * q.x - x * q.z,
		w * q.z + z * q.w + x * q.y - y * q.x
	);
}

GtQuaternion GtQuaternion::operator * (float fValue) const
{
	return GtQuaternion(fValue * w, fValue * x, fValue * y, fValue * z);
}

GtQuaternion GtQuaternion::operator += (const GtQuaternion& q)
{
	SetW(w + q.w);
	SetX(x + q.x);
	SetY(y + q.y);
	SetZ(z + q.z);
	return *this;
}

GtQuaternion GtQuaternion::operator -= (const GtQuaternion& q)
{
	SetW(w - q.w);
	SetX(x - q.x);
	SetY(y - q.y);
	SetZ(z - q.z);
	return *this;
}

GtQuaternion GtQuaternion::operator *= (const GtQuaternion& q)
{
	GtQuaternion result;
	result.SetW(w * q.w - x * q.x - y * q.y - z * q.z);
	result.SetX(w * q.x + x * q.w + y * q.z - z * q.y);
	result.SetY(w * q.y + y * q.w + z * q.x - x * q.z);
	result.SetZ(w * q.z + z * q.w + x * q.y - y * q.x);
	return result;
}

GtQuaternion GtQuaternion::operator *= (const float& fValue)
{
	w *= fValue;
	x *= fValue;
	y *= fValue;
	z *= fValue;
	return *this;
}

void GtQuaternion::ToAngleAxis(float& angle, GtPoint3f& axis) const
{
	float length = GtSqrt(x * x + y * y + z * z);
	angle = 2.0f/GtACos(w);
	float invLength = 1.0f/length;
	axis.x = x * invLength;
	axis.y = y * invLength;
	axis.z = z * invLength;
}

void GtQuaternion::FromAngleAxis(float angle, const GtPoint3f& axis)
{
	float fHalfAngle = angle * 0.5f;
	float sn;
	GtSinCos(fHalfAngle, sn, w);
	x = axis.x * sn;
	y = axis.y * sn;
	z = axis.z * sn;
}

void GtQuaternion::ToRotation(GtMatrix4& rot) const
{
	// operations ( * , + , - ) = 24

    float tx  = 2.0f*x;
    float ty  = 2.0f*y;
    float tz  = 2.0f*z;
    float twx = tx*w;
    float twy = ty*w;
    float twz = tz*w;
    float txx = tx*x;
    float txy = ty*x;
    float txz = tz*x;
    float tyy = ty*y;
    float tyz = tz*y;
    float tzz = tz*z;

	rot.SetCol(0, 1.0f-(tyy+tzz), txy+twz, txz-twy);
	rot.SetCol(1, txy-twz, 1.0f-(txx+tzz), tyz+twx);
	rot.SetCol(2, txz+twy, tyz-twx, 1.0f-(txx+tyy));
}

void GtQuaternion::FromRotation(const GtMatrix4& rot)
{
	float fTrace = rot.m[0][0] + rot.m[1][1] + rot.m[2][2];
	float fRoot;

	if(fTrace > 0.0f)
	{
		fRoot = GtSqrt(fTrace + 1.0f);
		w = 0.5f * fRoot;
		fRoot = 0.5f/fRoot;

		x = (rot.m[2][1] - rot.m[1][2]) * fRoot;
		y = (rot.m[0][2] - rot.m[2][0]) * fRoot;
		z = (rot.m[1][0] - rot.m[0][1]) * fRoot;
	}
	else
	{
		static int next[3] = { 1, 2, 0 };
		int i = 0;
		if( rot.m[1][1] > rot.m[0][0] )
			i = 1;
		if( rot.m[2][2] > rot.m[i][i] )
			i = 2;
		int j = next[i];
		int k = next[j];

		fRoot = GtSqrt(rot.m[i][i] - rot.m[j][j] - rot.m[k][k] + 1.0f);
		float* quat[3] = { &x, &y, &z };
		*quat[i] = 0.5f * fRoot;
		fRoot = 0.5f/fRoot;
		w = (rot.m[k][j] - rot.m[j][k]) * fRoot;
		*quat[j] = (rot.m[j][i] + rot.m[i][j]) * fRoot;
		*quat[k] = (rot.m[k][i] + rot.m[i][k]) * fRoot;
	}
}

void GtQuaternion::FromAngleAxisX(float fAngle)
{
	GtSinCos(fAngle * 0.5f, x, w);
	y = z = 0.0f;
}

void GtQuaternion::FromAngleAxisY(float fAngle)
{
	GtSinCos(fAngle * 0.5f, y, w);
	x = z = 0.0f;
}

void GtQuaternion::FromAngleAxisZ(float fAngle)
{
	GtSinCos(fAngle * 0.5f, z, w);
	x = y = 0.0f;
}

float GtQuaternion::Dot(const GtQuaternion &p, const GtQuaternion &q)
{
	return (p.w * q.w + p.x * q.x +
			p.y * q.y + p.z * q.z);
}

GtQuaternion GtQuaternion::UnitInverse(const GtQuaternion &q)
{
	return GtQuaternion(q.w, -q.x, -q.y, -q.z);
}

GtQuaternion GtQuaternion::Exp(const GtQuaternion& q)
{
	float fAngle = GtSqrt(q.x * q.x + q.y * q.y + q.z * q.z);
	float sn, cs;
	GtSinCos(fAngle, sn, cs);

	float coeff = (GtAbs(sn) < ms_fEpsilon ? 1.0f : sn/fAngle);

	GtQuaternion result(cs, coeff * q.x, coeff * q.y, coeff * q.z);

	return result;
}

GtQuaternion GtQuaternion::Log(const GtQuaternion& q)
{
	float fAngle = GtACos(q.w);
	float sn = GtSin(fAngle);

	float coeff = (GtAbs(sn) < ms_fEpsilon ? 1.0f : fAngle/sn );

	return GtQuaternion(0.0f, coeff * q.x, coeff * q.y, coeff * q.z);
}

GtQuaternion GtQuaternion::Slerp(float t, const GtQuaternion &p, const GtQuaternion &q)
{
	float fCos = Dot(p, q);

	float fTPrime;
	if(t<=0.5f)
		fTPrime = CounterWarp(t, fCos);
	else
		fTPrime = 1.0f - CounterWarp(1.0f - t, fCos);

	GtQuaternion result(
        Lerp(p.GetW(), q.GetW(), fTPrime),
        Lerp(p.GetX(), q.GetX(), fTPrime),
        Lerp(p.GetY(), q.GetY(), fTPrime),
        Lerp(p.GetZ(), q.GetZ(), fTPrime)
		);

	result.FastNormalize();

	return result;
}

GtQuaternion GtQuaternion::Intermediate(const GtQuaternion &q0, const GtQuaternion &q1, const GtQuaternion &q2)
{
	GtQuaternion inv = UnitInverse(q1);
	GtQuaternion exp = Exp((Log(inv * q0) + Log(inv * q2)) * -0.25f);

	return q1 * exp;
}

GtQuaternion GtQuaternion::Squad(float t, const GtQuaternion &p, const GtQuaternion &a, const GtQuaternion &b, const GtQuaternion &q)
{
	return Slerp(2.0f * t * (1.0f - t), Slerp(t, p, q), Slerp(t, a, b));
}

//---------------------------------------------------------------------------
//Ӻ긮 
static const float ISQRT_NEIGHBORHOOD = 0.959066f;
static const float ISQRT_SCALE = 1.000311f;
static const float ISQRT_ADDITIVE_CONSTANT = ISQRT_SCALE / 
    (float)sqrt(ISQRT_NEIGHBORHOOD);
static const float ISQRT_FACTOR = ISQRT_SCALE * (-0.5f / 
    (ISQRT_NEIGHBORHOOD * (float)sqrt(ISQRT_NEIGHBORHOOD)));
float GtQuaternion::ISqrt_approx_in_neighborhood(float s)
{
    return ISQRT_ADDITIVE_CONSTANT + (s - ISQRT_NEIGHBORHOOD) * ISQRT_FACTOR;
}

// Slerp  ȴ
float GtQuaternion::Lerp(float v0, float v1, float fPerc)
{
	return v0 + fPerc * ( v1 - v0 );
}

float GtQuaternion::CounterWarp(float t, float fCos)
{
    const float ATTENUATION = 0.82279687f;
    const float WORST_CASE_SLOPE = 0.58549219f;

    float fFactor = 1.0f - ATTENUATION * fCos;
    fFactor *= fFactor;
    float fK = WORST_CASE_SLOPE * fFactor;

    return t*(fK*t*(2.0f*t - 3.0f) + 1.0f + fK);
}

void GtQuaternion::Normalize(void)
{
	float fLength = w * w + x * x + y * y + z * z;
	float fInvLength = 1.0f/GtSqrt(fLength);
	*this = *this * fInvLength;
}

void GtQuaternion::FastNormalize()
{
	float s = x * x + y * y + z * z + w * w;
	float k = ISqrt_approx_in_neighborhood(s);

    if (s <= 0.91521198f) {
        k *= ISqrt_approx_in_neighborhood(k * k * s);

        if (s <= 0.65211970f) {
            k *= ISqrt_approx_in_neighborhood(k * k * s);
        }
    }

    x *= k;
    y *= k;
    z *= k;
    w *= k;
}